#############################################################################
##
##  forms.gi              'Forms' package
##                                                              John Bamberg
##                                                              Jan De Beule
##
##  Copyright 2007 Ghent University
##
##  Implementation of quadratic and sesquilinear forms
##
#############################################################################

#############################################################################
# Fundamental methods:
#############################################################################

InstallMethod( \=, "for two forms",
  [IsTrivialForm and IsFormRep, IsTrivialForm and IsFormRep],
  function( a, b )
    if a!.basefield <> b!.basefield then 
       return false;
    fi;
    return true;
  end );

InstallMethod( \=, "for two forms",
  [IsForm and IsFormRep, IsForm and IsFormRep],
  function( a, b )
    local aa,bb,p,s,i;
    if a!.basefield <> b!.basefield then 
       return false;
    fi;
    if a!.type <> b!.type then
       return false;
    fi;
    aa := a!.matrix;
    bb := b!.matrix;
    if Size(aa) <> Size(bb) then 
       return false;
    fi;
    p := PositionNonZero(aa[1]);
    s := bb[1][p] / aa[1][p];
    for i in [1..Length(aa)] do
        if s*aa[i] <> bb[i] then return false; fi;
    od;
    return true;
  end );

#############################################################################
# Form-by-matrix methods:
#############################################################################

InstallMethod( FormByMatrix, "for a ffe matrix, a field and a string",
  [IsMatrix and IsFFECollColl, IsField and IsFinite, IsString],
  function( m, f, string )
    local el;
    el := rec( matrix := m, basefield := f, type := string );

   ## We follow a certain convention, which is outlined in the manual,
   ## in order to determine from a Gram matrix the type of the constructed
   ## form.

    if string = "hermitian" then
      if not IsInt(Sqrt(Size(f))) then
        Error("No hermitian form exist when the order of <f> is not a square\n" );
      fi;
      if IsHermitianMatrix(m,f) then
        Objectify(NewType( HermitianFormFamily ,  IsFormRep),  el);
	return el;
      else
        Error("Given matrix does not define a hermitian form\n" );
      fi;   
    elif string = "symplectic" then
      if IsSymplecticMatrix(m,f) then
        Objectify(NewType( BilinearFormFamily ,  IsFormRep),  el);
	return el;
      else
        Error("Given matrix does not define a symplectic form\n" );
      fi;
    elif string = "orthogonal" then
      if IsEvenInt(Size(f)) then
        Error("No orthogonal forms exist in even characteristic\n" );
      fi;
      if IsOrthogonalMatrix(m) then
        Objectify(NewType( BilinearFormFamily ,  IsFormRep),  el);
	return el;
      else
        Error("Given matrix does not define an orthogonal form\n" );
      fi;
    elif string = "quadratic" then
      if IsOddInt(Size(f)) then
        Error("Please specify a quadratic form in odd characteristic as an orthogonal form\n" );
      else
        el.matrix := RESET(m,Length(m),Size(f));
	Objectify(NewType( QuadraticFormFamily ,  IsFormRep),  el);
        return el;
      fi;
    else
      Error("Please specify a form properly\n");
    fi;
  end );

InstallMethod( BilinearFormByMatrix, "for a ffe matrix and a field",
  [IsMatrix and IsFFECollColl, IsField and IsFinite],
  function( m, f )
    local el;
    if IsZero(m) then
       el := rec( matrix := m, basefield := f, type := "trivial" );
       Objectify(NewType( TrivialFormFamily ,  IsFormRep),  el);
 	return el;
    elif IsSymplecticMatrix(m,f) then
       el := rec( matrix := m, basefield := f, type := "symplectic" );
       Objectify(NewType( BilinearFormFamily ,  IsFormRep),  el);
	return el;
    elif IsOrthogonalMatrix(m) then 
       el := rec( matrix := m, basefield := f, type := "orthogonal" );
       Objectify(NewType( BilinearFormFamily ,  IsFormRep),  el);
 	return el;
    else
       Error("Invalid Gram matrix\n");
    fi;
  end );

InstallMethod( QuadraticFormByMatrix, "for a ffe matrix and a field",
  [IsMatrix and IsFFECollColl, IsField and IsFinite],
  function( m, f )
    local el;
    el := rec( matrix := m, basefield := f, type := "quadratic" );
    if IsZero(m) then
       Objectify(NewType( TrivialFormFamily ,  IsFormRep),  el);
 	return el;
    elif IsOddInt(Size(f)) then
       Objectify(NewType( QuadraticFormFamily,  IsFormRep),  el);
       return el;
    else
       el.matrix := RESET(m,Length(m),Size(f));
	Objectify(NewType( QuadraticFormFamily ,  IsFormRep),  el);
       return el;
    fi;
  end );

InstallMethod( HermitianFormByMatrix, "for a ffe matrix and a field",
  [IsMatrix and IsFFECollColl, IsField and IsFinite],
  function( m, f )
    local el;    
    if not IsInt(Sqrt(Size(f))) then
        Error("No hermitian form exist when the order of <f> is not a square\n" );
    fi;
    if IsHermitianMatrix(m,f) then
       el := rec( matrix := m, basefield := f, type := "hermitian" );
       Objectify(NewType( HermitianFormFamily ,  IsFormRep),  el);
       SetWittIndex( el, Int( (Length(m)+1)/2 ) );
	return el;
    else
       Error("Given matrix does not define a hermitian form\n" );
    fi;   
  end );

#############################################################################
# Helping functions for FormByPolynomial
#############################################################################

InstallMethod( GramMatrixByPolynomialForQuadraticForm, 
                      [IsPolynomial, IsField, IsInt, IsList],
  function(poly,gf,n,varlist)  
    local vars,q,polarity,mat,i,j,vals,der;
    q := Size(gf);
    mat := NullMat(n,n)*Z(q);
    vars := ShallowCopy(varlist);
    if IsZero(poly) then
      return mat;
    fi;
    n := Length(varlist);
    vals := [];
    for i in [1..n] do
      vals[i] := 0;
    od;
    for i in [1..n] do
      vals[i] := 1;
      mat[i][i] := Value(poly,vars,vals);
      vals[i] := 0;
    od;
    for i in [1..n-1] do
      der := Derivative(poly,vars[i]);
      for j in [i+1..n] do
        vals[j] := 1;
        mat[i][j] := Value(der,vars,vals);
        vals[j] := 0;
      od;
    od;
    vars := ShallowCopy(varlist);
    if vars*mat*vars <> poly then
      Error( "<poly> should be a homogeneous polynomial over GF(q)");
    fi; 
    if IsOddInt(q) then
      polarity := mat + TransposedMat(mat);
      polarity := polarity/2;
      return polarity;
    else
      return mat;
    fi;
end);

InstallMethod( GramMatrixByPolynomialForHermitianForm, 
                     [IsPolynomial, IsField, IsInt, IsList],
  function(poly,gf,n,varlist)
    local vars,varst,q,t,a,polarity,i,j,vals,der;
    q := Size(gf);
    t := Sqrt(q);
    vars := ShallowCopy(varlist);
    varst := List(vars,x->x^t);
    polarity := NullMat(n,n)*Z(q)^0;
    if IsZero(poly) then
      return polarity;
    fi;
    n := Length(varlist);
    vals := [];
    for i in [1..n] do
      vals[i] := 0;
    od;
    for i in [1..n] do
      vals[i] := 1;
      a := Value(poly,vars,vals);
      if a <> a^t then
        Error( "<poly> does not generate a Hermitian matrix" );
      else
        polarity[i][i] := a;
      fi;
      vals[i] := 0;
    od;
    for i in [1..n-1] do
      der := Derivative(poly,vars[i]);
      for j in [i+1..n] do
        vals[j] := 1;
        polarity[i][j] := Value(der,vars,vals);
        polarity[j][i] := polarity[i][j]^t;
        vals[j] := 0;
      od;
    od;
    vars := ShallowCopy(varlist);
      #Check whether the polynomial has the right form.
    if vars*polarity*varst <> poly then
      Error( "<poly> does not generate a Hermitian matrix" );
    fi;
  return polarity;
end );

#############################################################################
# Form-by-polynomial methods
#############################################################################
  
InstallMethod( FormByPolynomial,
  "for a polynomial over a finite field and a string",
  [IsPolynomial, IsField and IsFinite, IsInt, IsList, IsString],
  function(pol,gf,n,vars,string)
  local mat, form;
  if string = "orthogonal" then
    if IsEvenInt(Size(gf)) then
      Error("No orthogonal forms exist in even characteristic\n");
    else
      mat := GramMatrixByPolynomialForQuadraticForm(pol,gf,n,vars);
      form := FormByMatrix(mat,gf,"orthogonal");
      SetPolynomialOfForm(form, pol);
      return form;
    fi;
  elif string = "quadratic" then
    if IsOddInt(Size(gf)) then
      Error("Please specify quadratic forms in odd characteristic as orthogonal forms\n");
    else
      mat := GramMatrixByPolynomialForQuadraticForm(pol,gf,n,vars);
      form := FormByMatrix(mat,gf,"quadratic");
      SetPolynomialOfForm(form, pol);
      return form;
    fi;
  elif string = "hermitian" then
    mat := GramMatrixByPolynomialForHermitianForm(pol,gf,n,vars);
    form := FormByMatrix(mat,gf,"hermitian");
    SetPolynomialOfForm(form, pol);
    return form;
  else
    Error("No other forms than quadratic or hermitian can be specified by polynomials\n"); 
  fi;
  end );

InstallMethod( FormByPolynomial,
  "for a polynomial over a finite field and a string",
  [IsPolynomial, IsFiniteFieldPolynomialRing, IsInt, IsString],
  function(pol,pring,n,string)
  local mat, form, gf, vars;
  gf := CoefficientsRing( pring );
  vars := IndeterminatesOfPolynomialRing( pring );
  if string = "orthogonal" then
    if IsEvenInt(Size(gf)) then
      Error("No orthogonal forms exist in even characteristic\n");
    else
      mat := GramMatrixByPolynomialForQuadraticForm(pol,gf,n,vars);
      form := FormByMatrix(mat,gf,"orthogonal");
      SetPolynomialOfForm(form, pol);
      return form;
    fi;
  elif string = "quadratic" then
    if IsOddInt(Size(gf)) then
      Error("Please specify quadratic forms in odd characteristic as orthogonal forms\n");
    else
      mat := GramMatrixByPolynomialForQuadraticForm(pol,gf,n,vars);
      form := FormByMatrix(mat,gf,"quadratic");
      SetPolynomialOfForm(form, pol);
      return form;
    fi;
  elif string = "hermitian" then
    mat := GramMatrixByPolynomialForHermitianForm(pol,gf,n,vars);
    form := FormByMatrix(mat,gf,"hermitian");
    SetPolynomialOfForm(form, pol);
    return form;
  else
    Error("No other forms than quadratic or hermitian can be specified by polynomials\n"); 
  fi;
  end );

InstallMethod( BilinearFormByPolynomial,
  "for a polynomial over a field, and a dimension",
  [IsPolynomial, IsFiniteFieldPolynomialRing, IsInt],
  function( pol, pring, n )
  local mat, form, gf, vars;
  gf := CoefficientsRing( pring );
  vars := IndeterminatesOfPolynomialRing( pring );
  if IsEvenInt(Size(gf)) then
      Error("No orthogonal form exists in even characteristic\n");
  else
      mat := GramMatrixByPolynomialForQuadraticForm(pol,gf,n,vars);
      form := FormByMatrix(mat,gf,"orthogonal");
      SetPolynomialOfForm(form, pol);
      return form;
  fi;
  end );

InstallMethod( BilinearFormByPolynomial,  "no dimension",
  [IsPolynomial, IsFiniteFieldPolynomialRing],
  function( pol, pring )
  local n;
    n := Length(IndeterminatesOfPolynomialRing( pring ));
    return BilinearFormByPolynomial( pol, pring, n);
  end );

InstallMethod( HermitianFormByPolynomial,
  "for a polynomial over a field, and a dimension",
  [IsPolynomial, IsFiniteFieldPolynomialRing, IsInt],
  function(pol, pring, n)
  local mat, form, gf, vars;
  gf := CoefficientsRing( pring );
  vars := IndeterminatesOfPolynomialRing( pring );
  mat := GramMatrixByPolynomialForHermitianForm(pol,gf,n,vars);
  form := FormByMatrix(mat,gf,"hermitian");
  SetPolynomialOfForm(form, pol);
  return form;
  end );

InstallMethod( HermitianFormByPolynomial,  "no dimension",
  [IsPolynomial, IsFiniteFieldPolynomialRing],
  function( pol, pring )
  local n;
    n := Length(IndeterminatesOfPolynomialRing( pring ));
    return HermitianFormByPolynomial( pol, pring, n);
  end );

InstallMethod( QuadraticFormByPolynomial,
  "for a polynomial over a field, and a dimension",
  [IsPolynomial, IsFiniteFieldPolynomialRing, IsInt],
  function(pol, pring, n)
  local mat, form, gf, vars;
  gf := CoefficientsRing( pring );
  vars := IndeterminatesOfPolynomialRing( pring );
  if IsOddInt(Size(gf)) then
     Error("Please specify quadratic forms in odd characteristic as orthogonal forms\n");
  else
     mat := GramMatrixByPolynomialForQuadraticForm(pol,gf,n,vars);
     form := FormByMatrix(mat,gf,"quadratic");
     SetPolynomialOfForm(form, pol);
     return form;
  fi;
  end );

InstallMethod( QuadraticFormByPolynomial,  "no dimension",
  [IsPolynomial, IsFiniteFieldPolynomialRing],
  function( pol, pring )
  local n;
    n := Length(IndeterminatesOfPolynomialRing( pring ));
    return QuadraticFormByPolynomial( pol, pring, n);
  end );

InstallMethod( PolynomialOfForm, [IsQuadraticForm],
  function(f)
    ## This method finds the polynomial associated to
    ## the Gram matrix of f.

    local m, gf, d, r, indets, poly;
    m := f!.matrix;
    gf := f!.basefield; 
    d := Length(m);
    r := PolynomialRing(gf,d);
    indets := IndeterminatesOfPolynomialRing(r);
    poly := indets * m * indets;
    return poly;
  end );

InstallMethod( PolynomialOfForm, [IsHermitianForm],
  function(f)
    ## This method finds the polynomial associated to
    ## the Gram matrix of f.

    local m, gf, d, r, indets, poly, q;
    gf := f!.basefield; 
    q := Sqrt(Size(gf));
    m := f!.matrix;
    d := Length(m);
    r := PolynomialRing(gf,d);
    indets := IndeterminatesOfPolynomialRing(r);
    poly := indets * m * List(indets,t->t^q);
    return poly;
  end );


#############################################################################
# Base change methods
#############################################################################

InstallMethod( BaseChangeToCanonical, "for a sesquilinear form",
  [IsSesquilinearForm and IsFormRep],
  function(f)
    local string,m,n,q,gf,b;
    string := f!.type;
    m := f!.matrix;
    n := Length(m)-1;
    gf := f!.basefield;
    q := Size(gf);
    if string = "orthogonal" then
      b := BaseChangeOrthogonalBilinear(m, gf);
      if not HasWittIndex(f) then
         SetWittIndex(f, (b[2]+b[3]-1) / 2);
      fi;
      return b[1];
    elif string = "symplectic" then
      if not HasWittIndex(f) then
         SetWittIndex(f, (n+1)/2);
      fi;
      return BaseChangeSymplectic(m, gf);
    elif string = "hermitian" then
      b := BaseChangeHermitian(m, gf);
      if not HasWittIndex(f) then
         SetWittIndex(f, Int( (b[2]+1)/2 ));
      fi;
      return b[1];
    fi;
end );

InstallMethod( BaseChangeToCanonical, "for a quadratic form",
  [IsQuadraticForm and IsFormRep],
  function(f)
    local m, gf, b;
    m := f!.matrix;
    gf := f!.basefield;
    b := BaseChangeOrthogonalQuadratic(m, gf);
    if not HasWittIndex(f) then
         SetWittIndex(f, (b[2]+b[3]-1) / 2);
    fi;
    return b[1];
  end );

#############################################################################
# Overloading: Frobenius Automorphisms
#############################################################################

InstallOtherMethod( \^, "for a FFE vector and a Frobenius automorphism",
  [ IsVector and IsFFECollection, IsFrobeniusAutomorphism ],
  function( v, f )
    return List(v,x->x^f);
  end );

InstallOtherMethod( \^, "for a FFE vector and a trivial Frobenius automorphism",
  [ IsVector and IsFFECollection, IsMapping and IsOne ],
  function( v, f )
    return v;
  end );

InstallOtherMethod( \^, 
  "for a compressed GF2 vector and a Frobenius automorphism",
  [ IsVector and IsFFECollection and IsGF2VectorRep, IsFrobeniusAutomorphism ],
  function( v, f )
    local w;
    w := List(v,x->x^f);
    ConvertToVectorRepNC(w,2);
    return w;
  end );

InstallOtherMethod( \^, 
  "for a compressed GF2 vector and a trivial Frobenius automorphism",
  [ IsVector and IsFFECollection and IsGF2VectorRep, IsMapping and IsOne ],
  function( v, f )
    return v;
  end );

InstallOtherMethod( \^, 
  "for a compressed 8bit vector and a Frobenius automorphism",
  [ IsVector and IsFFECollection and Is8BitVectorRep, IsFrobeniusAutomorphism ],
  function( v, f )
    local w;
    w := List(v,x->x^f);
    ConvertToVectorRepNC(w,Q_VEC8BIT(v));
    return w;
  end );

InstallOtherMethod( \^, 
  "for a compressed 8bit vector and a trivial Frobenius automorphism",
  [ IsVector and IsFFECollection and Is8BitVectorRep, IsMapping and IsOne ],
  function( v, f )
    return v;
  end );

InstallOtherMethod( \^, "for a FFE matrix and a Frobenius automorphism",
  [ IsMatrix and IsFFECollColl, IsFrobeniusAutomorphism ],
  function( m, f )
    return List(m,v->List(v,x->x^f));
  end );

InstallOtherMethod( \^, "for a FFE matrix and a trivial Frobenius automorphism",
  [ IsMatrix and IsFFECollColl, IsMapping and IsOne ],
  function( m, f )
    return m;
  end );

InstallOtherMethod( \^, 
  "for a compressed GF2 matrix and a Frobenius automorphism",
  [ IsMatrix and IsFFECollColl and IsGF2MatrixRep, IsFrobeniusAutomorphism ],
  function( m, f )
    local w,l,i;
    l := [];
    for i in [1..Length(m)] do
        w := List(m[i],x->x^f);
        ConvertToVectorRepNC(w,2);
        Add(l,w);
    od;
    ConvertToMatrixRepNC(l,2);
    return l;
  end );

InstallOtherMethod( \^, 
  "for a compressed GF2 matrix and a trivial Frobenius automorphism",
  [ IsMatrix and IsFFECollColl and IsGF2MatrixRep, IsMapping and IsOne ],
  function( m, f )
    return m;
  end );

InstallOtherMethod( \^, 
  "for a compressed 8bit matrix and a Frobenius automorphism",
  [ IsMatrix and IsFFECollColl and Is8BitMatrixRep, IsFrobeniusAutomorphism ],
  function( m, f )
    local w,l,i,q;
    l := [];
    q := Q_VEC8BIT(m[1]);
    for i in [1..Length(m)] do
        w := List(m[i],x->x^f);
        ConvertToVectorRepNC(w,q);
        Add(l,w);
    od;
    ConvertToMatrixRepNC(l,q);
    return l;
  end );

InstallOtherMethod( \^, 
  "for a compressed 8bit matrix and a trivial Frobenius automorphism",
  [ IsMatrix and IsFFECollColl and Is8BitMatrixRep, IsMapping and IsOne ],
  function( m, f )
    return m;
  end );

#############################################################################
# Overloading: Forms
#############################################################################

InstallOtherMethod( \^, "for a pair of FFE vectors and a sesquilinear form",
  [ IsVectorList and IsFFECollColl, IsBilinearForm ],
  function( pair, f )
    if Size(pair) <> 2 then 
       Error("The first argument must be a pair of vectors");
    fi;
    return pair[1] * f!.matrix * pair[2];
  end );

InstallOtherMethod( \^, "for a pair of FFE matrices and a sesquilinear form",
  [ IsFFECollCollColl, IsBilinearForm ],
  function( pair, f )
    if Size(pair) <> 2 then 
       Error("The first argument must be a pair of vectors");
    fi;
    return pair[1] * f!.matrix * TransposedMat(pair[2]);
  end );

InstallOtherMethod( \^, "for a pair of FFE vectors and an hermitian form",
  [ IsVectorList and IsFFECollColl, IsHermitianForm ],
  function( pair, f )
    local frob;
    if Size(pair) <> 2 then 
       Error("The first argument must be a pair of vectors");
    fi;
    frob := FrobeniusAutomorphism(f!.basefield);
    return pair[1] * f!.matrix * (pair[2]^frob);
  end );

InstallOtherMethod( \^, "for a pair of FFE matrices and an hermitian form",
  [ IsFFECollCollColl, IsHermitianForm ],
  function( pair, f )
    local frob;
    if Size(pair) <> 2 then 
       Error("The first argument must be a pair of vectors");
    fi;
    frob := FrobeniusAutomorphism(f!.basefield);
    return pair[1] * f!.matrix * (TransposedMat(pair[2])^frob);
  end );

InstallOtherMethod( \^, "for a FFE vector and a quadratic form",
  [ IsVector and IsFFECollection, IsQuadraticForm ],
  function( v, f )
    return v * f!.matrix * v;
  end );

InstallOtherMethod( \^, "for a FFE matrix and a quadratic form",
  [ IsMatrix and IsFFECollColl, IsQuadraticForm ],
  function( m, f )
    return m * f!.matrix * TransposedMat(m);
  end );

#############################################################################
# Viewing methods:
#############################################################################

InstallMethod( ViewObj, [ IsTrivialForm ],
  function( f )
    Print("< trivial form >");
  end );

InstallMethod( PrintObj, [ IsTrivialForm ],
  function( f )
    Print("Trivial form\n");
    Print("Gram Matrix:\n",f!.matrix,"\n");
  end );

InstallMethod( Display, [ IsTrivialForm ],
  function( f )
    Print("Trivial form\n");
    Print("Gram Matrix:\n");
    Display(f!.matrix);
  end);

InstallMethod( ViewObj, [ IsHermitianForm ],
  function( f )
    Print("< hermitian form >");
  end );

InstallMethod( PrintObj, [ IsHermitianForm ],
  function( f )
    Print("Hermitian form\n");
    Print("Gram Matrix:\n",f!.matrix,"\n");
    if HasPolynomialOfForm( f ) then
       Print("Polynomial: ", PolynomialOfForm, "\n");
    fi;
    if HasWittIndex( f ) then 
       Print("Witt Index: ", WittIndex(f), "\n");
    fi;        
  end );

InstallMethod( Display, [ IsHermitianForm ],
  function( f )
    Print("Hermitian form\n");
    Print("Gram Matrix:\n");
    Display(f!.matrix);
    if HasPolynomialOfForm( f ) then
       Print("Polynomial: "); 
       Display(PolynomialOfForm(f));
       Print("\n");
    fi;
    if HasWittIndex( f ) then 
       Print("Witt Index: ", WittIndex(f), "\n");
    fi;
  end);

InstallMethod( ViewObj, [ IsQuadraticForm ],
  function( f )
    if HasIsEllipticForm( f ) or HasIsHyperbolicForm( f ) or
       HasIsParabolicForm( f ) then
       if IsEllipticForm( f ) then
          Print("< elliptic quadratic form >");
       elif IsHyperbolicForm( f ) then
          Print("< hyperbolic quadratic form >");
       elif IsParabolicForm( f)  then
          Print("< parabolic quadratic form >");
       else 
          Print("< quadratic form >");
       fi;
    else
      Print("< quadratic form >");
    fi;
  end );

InstallMethod( PrintObj, [ IsQuadraticForm ],
  function( f )
    if HasIsEllipticForm( f ) or HasIsHyperbolicForm( f ) or
       HasIsParabolicForm( f ) then
       if IsEllipticForm( f ) then
          Print("Elliptic quadratic form\n");
       elif IsHyperbolicForm( f ) then
          Print("Hyperbolic quadratic form\n");
       elif IsParabolicForm( f)  then
          Print("Parabolic quadratic form\n");
       else 
          Print("Quadratic form\n");
       fi;
    else
      Print("Quadratic form\n");
    fi;
    Print("Gram Matrix:\n",f!.matrix,"\n");
    if HasPolynomialOfForm( f ) then
       Print("Polynomial: ", PolynomialOfForm, "\n");
    fi;
    if HasWittIndex( f ) then 
       Print("Witt Index: ", WittIndex(f), "\n");
    fi;
  end );

InstallMethod( Display,  [ IsQuadraticForm ],
  function( f )
    if HasIsEllipticForm( f ) or HasIsHyperbolicForm( f ) or
       HasIsParabolicForm( f ) then
       if IsEllipticForm( f ) then
          Print("Elliptic Quadratic form\n");          
       elif IsHyperbolicForm( f ) then
          Print("Hyperbolic Quadratic form\n");
       elif IsParabolicForm( f)  then
          Print("Parabolic Quadratic form\n");
       else 
          Print("Quadratic form\n");
       fi;
    else
      Print("Quadratic form\n");
    fi;
    Print("Gram Matrix:\n");
    Display(f!.matrix);
    if HasPolynomialOfForm( f ) then
       Print("Polynomial: "); 
       Display(PolynomialOfForm(f));
       Print("\n");
    fi;
    if HasWittIndex( f ) then 
       Print("Witt Index: ", WittIndex(f), "\n");
    fi;
  end );

InstallMethod( ViewObj, [ IsBilinearForm ],
  function( f )
    if HasIsEllipticForm( f ) or HasIsHyperbolicForm( f ) or
       HasIsParabolicForm( f ) then
       if IsEllipticForm( f ) then
          Print("< elliptic bilinear form >");
       elif IsHyperbolicForm( f ) then
          Print("< hyperbolic bilinear form >");
       elif IsParabolicForm( f)  then
          Print("< parabolic bilinear form >");
       else 
          Print("< bilinear form >");
       fi;
    else
      Print("< bilinear form >");
    fi;
  end );

InstallMethod( PrintObj, [ IsBilinearForm ],
  function( f )
    if HasIsEllipticForm( f ) or HasIsHyperbolicForm( f ) or
       HasIsParabolicForm( f ) then
       if IsEllipticForm( f ) then
          Print("Elliptic bilinear form\n");
       elif IsHyperbolicForm( f ) then
          Print("Hyperbolic bilinear form\n");
       elif IsParabolicForm( f)  then
          Print("Parabolic bilinear form\n");
       else 
          Print("Bilinear form\n");
       fi;
    else
      Print("Bilinear form\n");
    fi;
    Print("Gram Matrix:\n",f!.matrix,"\n");
    if HasPolynomialOfForm( f ) then
       Print("Polynomial: ", PolynomialOfForm, "\n");
    fi;
    if HasWittIndex( f ) then 
       Print("Witt Index: ", WittIndex(f), "\n");
    fi;
  end );

InstallMethod( Display, [ IsBilinearForm ],
  function( f )
    if HasIsEllipticForm( f ) or HasIsHyperbolicForm( f ) or
       HasIsParabolicForm( f ) then
       if IsEllipticForm( f ) then
          Print("Elliptic bilinear form\n");          
       elif IsHyperbolicForm( f ) then
          Print("Hyperbolic bilinear form\n");
       elif IsParabolicForm( f)  then
          Print("Parabolic bilinear form\n");
       else 
          Print("Bilinear form\n");
       fi;
    else
      Print("Bilinear form\n");
    fi;
    Print("Gram Matrix:\n");
    Display(f!.matrix);
    if HasPolynomialOfForm( f ) then
       Print("Polynomial: "); 
       Display(PolynomialOfForm(f));
       Print("\n");
    fi;
    if HasWittIndex( f ) then 
       Print("Witt Index: ", WittIndex(f), "\n");
    fi;
  end );

#############################################################################
# Functions:
#############################################################################
InstallGlobalFunction(SWR,
  function(i,j,n)
    local P;
    P := IdentityMat(n);
    P[i][i] := 0;
    P[i][j] := 1;
    P[j][j] := 0;
    P[j][i] := 1;
    return P;
end );

InstallGlobalFunction(SUM_OF_SQUARES,
    function(v,q)
    local stop,dummy,i,v1,v2, primroot;
    primroot := Z(q);
    stop := false;
    i := 0;
    repeat
      dummy := LogFFE(v - primroot^(2*i), primroot);
      if dummy mod 2 = 0 then
        v1 := primroot^i;
        v2 := primroot^(dummy/2);
        stop := true;
      else
        i := i + 1;
      fi;  
    until stop;
    return [v1,v2];
end );


InstallGlobalFunction(REDUCE2,
  function(start,stop,n,q)
    local t,i,P,one,primroot;
    primroot := Z(q);
    one := primroot^0;
    t := primroot^(LogFFE(-one,primroot)/2);
    P := IdentityMat(n,GF(q));
    i := start;
    while i < stop do
      P[i][i] := (1/2)*one;
      P[i][i+1] := (1/2)*t;
      P[i+1][i] := (1/2)*one;
      P[i+1][i+1] := (-1/2)*t;
      i := i + 2;
    od;
    return P;
end );

InstallGlobalFunction(REDUCE4,
    function(start,stop,n,q)
    local c,d,i,P,dummy;
    P := IdentityMat(n,GF(q));
    i := start;
    dummy := SUM_OF_SQUARES(-One(GF(q)),q);
    c := dummy[1];
    d := dummy[2];
    while i < stop do
      P[i+1][i+1] := c;
      P[i+1][i+3] := d;
      P[i+3][i+1] := d;
      P[i+3][i+3] := -c;
      i := i + 4;
    od;
    return P;
end );

InstallGlobalFunction(DIFF_2_S,
    function(start,stop,n,q)
    local i,P,one;
    i := start;
    P := IdentityMat(n,GF(q)); 
    one := Z(q)^0;
    while i < stop do
      P[i][i] := one / 2;
      P[i][i+1] := one / 2;
      P[i+1][i] := one / 2;
      P[i+1][i+1] := -one / 2;
      i := i + 2;
    od;
    return P;
end );

InstallGlobalFunction(HERM_CONJ,
    function(mat,n,t)
    local i,j,dummy;
    dummy := MutableTransposedMat(mat);
    for i in  [1..n] do
      for j in [1..n] do
        dummy[i][j] := dummy[i][j]^t;
      od;
    od;
    return dummy;
end );

InstallGlobalFunction(RESET,
    function(mat,n,q)
    local i,j,A,B,t;
    t := 0*Z(q);
    A := List(mat,ShallowCopy);
    B := List(TransposedMat(A),ShallowCopy);
    for i in [1..n] do
      B[i][i] := t;
    od;
    A := A + B;
    for i in [2..n] do  
      for j in [1..i-1] do 
        A[i][j] := t;
      od;
    od;
    return A;
end);

InstallGlobalFunction(SQRT2,
    function(a, q)
    if IsZero( a ) then
      return a;
    else
      return Z(q)^((q*LogFFE(a,Z(q)))/2);
    fi;
end);

InstallGlobalFunction(PERM_VAR,
    function(n,r)
    local i,P;
    P := IdentityMat(n);
    for i in [1..r-1] do
      P[i][i] := 0;
      P[i+1][i] := 1;
    od;
    P[r][r] := 0;
    P[1][r] := 1;
    return P;
end);

InstallGlobalFunction(C1,
  function(q, h)
    local i, primroot;
    primroot := Z(q);
    if h mod 2 = 0 then
      i := 1;
      while i <= h - 1 do
        if not IsZero( Trace(GF(q), primroot^i) ) then
          return primroot^i;
        else
          i := i + 1;
        fi;
      od;
    else
      return primroot^0;
    fi;
end );

InstallGlobalFunction(QUAD_EQ,
    function(delta,q,h)
    local i,k,dummy,result;
    k := C1(q,h);
    dummy := 0*Z(q);
    result := 0*Z(q);
    for i in [1..h-1] do
      dummy := dummy + k^(2^(i-1));
      result := result + dummy*(delta^(2^i));
    od;
    return result;
end);


#############################################################################
# Functions (most likely not for the user):
#############################################################################

InstallMethod( IsSymplecticMatrix, [IsFFECollColl, IsField],
  function(m,f)
    local n,bool,i;
    n := Length(m);
    bool := true;
    if IsEvenInt(Size(f)) then
      bool := ForAll([1..n],i->IsZero(m[i][i]));
    fi;
    return m = -TransposedMat(m) and bool;
  end );

InstallMethod( IsOrthogonalMatrix, [IsFFECollColl],
  function(m)
    return m=TransposedMat(m);
  end );

InstallMethod( IsHermitianMatrix, [IsFFECollColl, IsField],
  function(m,f)
    local t,n; 
    t := Sqrt(Size(f));
    n := Length(m);
    return m=HERM_CONJ(m,n,t);
  end );

#############################################################################
# Main Base-change operations:
#############################################################################

InstallMethod( BaseChangeOrthogonalBilinear, 
    [ IsMatrix and IsFFECollColl, IsField and IsFinite ],
  function(mat, gf)
    local row,i,j,k,A,b,c,d,P,D,dummy,r,w,s,v,v1,v2,
          stop,stop2,nplus1,nplus2,q,primroot,one,n;
    A := ShallowCopy(mat);
    #  n is the projective dimension
    nplus1 := Size(mat);
    n := nplus1 - 1;
    nplus2 := n + 2;
    q := Size(gf);
    D := IdentityMat(nplus1, gf);
    row := 0;
    stop := false;
    primroot := Z(q);
    one := One(GF(q)); 

    repeat       

    # We search for an element different from zero on
    # the main diagonal from row + 1 onwards

      i := 1 + row;
      dummy := false;
      while dummy = false and i <= nplus1 do
        if IsZero( A[i][i] ) then
           i := i + 1;
        else
           dummy := true;
        fi;
      od;

      # if i is row + 1, then we do nothing since A[i][i] <> 0
      if 2 + row <= i and i <= nplus1 then
        P := SWR(row + 1,i,nplus1);
        A := P*A*P;
        D := P*D;

      # If we have not found it on the main diagonal, then we go and do other work.
      # We search for a nonzero element off the main diagonal.

      elif i = nplus2 then
        i := 1 + row;
        dummy := false;
        while i <= n and dummy = false do
          k := i + 1; 
          while k <= nplus1 and dummy = false do
            if IsZero(A[i][k]) then
               k := k + 1;
            else
               dummy := true;
            fi;
          od;
          if k = nplus2 then
             i := i + 1;
          fi;
        od;

        # if i is n+1, then they are all zero and we can stop.

        if i = nplus1 then
          stop := true;
          r := row;

        # Otherwise: Go and fetch...
        # Put it on A[row+1,row+2]

        elif i = row + 1 then 
          P := SWR(row+2,k,nplus1);
          A := P*A*P;
          D := P*D;
        else
          P := SWR(row+2,k,nplus1) * SWR(row+1,i,nplus1);
          A := P*A*TransposedMat(P);
          D := P*D;
        fi;

        #...take care that there is a nonzero on the main diagonal

        if not stop then
          b := A[row+2][row+1]^-1;
          P := IdentityMat(nplus1, gf);
          P[row+1][row+2] := b;
          A := P*A*TransposedMat(P);
          D := P*D;
        fi;
      fi;   # end if 2+row <= i and i <= nplus1 ... elif ... fi

      # There is no zero element on the main diagonal, make the rest zero.

      if not stop then
        P := IdentityMat(nplus1, gf);
        for i in [row+2..nplus1] do
           P[i][row+1] := -A[i][row+1] * A[row+1][row+1]^-1;   ## bug here!
        od;
        A := P*A*TransposedMat(P);
        D := P*D;
        row := row + 1;
      fi;
    until row = n or stop;
   
    # Count how many variables are used.    

    if not stop then
      if not IsZero( A[nplus1][nplus1] ) then
        r := nplus1;
      else
        r := n;
      fi;
    fi;

    # Here we distinguish the quadratic from the non-quadratic

    i := 1;
    s := 0;
    stop := false;
    while (not stop) and i < r do
      if IsOddInt( LogFFE(A[i][i], primroot) ) then
         j := i + 1;
         stop2 := false;
         repeat
            if IsEvenInt( LogFFE(A[j][j], primroot) ) then
               dummy := A[j][j];
               A[j][j] := A[i][i];
               A[i][i] := dummy;
               D := SWR(i,j,nplus1)*D;
               stop2 := true;
               i := i + 1;
               s := s + 1;
            else
               j := j + 1;
            fi;
         until stop2 or j = r + 1;
         if j = r + 1 then
            stop := true;
         fi;
      else
        i := i + 1;
        s := s + 1;
      fi;
    od; 
    if IsEvenInt( LogFFE(A[r][r], primroot) ) then
       s := s + 1;
    fi; 

    # We do the form x_0^2 + ... + x_s^2 + v(x_s+1^2 + ... + x_r^2)
    # with v not quadratic.

    P := IdentityMat(nplus1, gf);
    v := ShallowCopy(primroot);
    for i in [1..s] do
      P[i][i] := (primroot^(LogFFE(A[i][i], primroot)/2))^-1;
    od;
    for i in [s+1..r] do
      P[i][i] := (primroot^(LogFFE(A[i][i]/primroot,primroot)/2))^-1;
    od;
    D := P*D;

    # We keep as much quadratic part as we can:
    
    s := s - 1;
    r := r - 1;
    
    # We write first v=v1^2 + v2^2
    
    dummy := SUM_OF_SQUARES(v,q);
    v1 := dummy[1];
    v2 := dummy[2];

    # Case by case:

    P := IdentityMat(nplus1, gf);
    if not (s = -1 or r = s )  then
      if (r - s) mod 2 = 0 then
        for i in [s+2..r+1] do
          P[i][i] := v1/v;
        od;
        i := s + 2;
        repeat
          P[i][i+1] := -v2/v;
          P[i+1][i] := v2/v;
          i := i + 2;
        until i = r + 2;
        D := P*D;
        s := r;
      else
        if r mod 2 = 0 then
          for i in [1..s+1] do
            P[i][i] := v1;
          od;
          i := 1;
          repeat
            P[i][i+1] := v2;
            P[i+1][i] := -v2;
            i := i + 2;
          until i = s + 2;
          D := P*D;
          s := -1;
        elif not (s = r - 1) then
          for i in [s+2..r] do
            P[i][i] := v1/v;
          od;
          i := s + 2;
          repeat
            P[i][i+1] := -v2/v;
            P[i+1][i] := v2/v;
            i := i + 2;
          until i = r + 1;
          D := P*D;
          s := r - 1;
        fi; 
      fi; 
    fi; 
    
    # Towards standard forms (uses the standard proof of
    # the classification of quadratic forms):

    if r mod 2 <> 0 then
      if s = -1 or s = r then
        if q mod 4 = 1 then
          P := REDUCE2(1,r+1,nplus1,q);
          D := P*D;
          w := 2;
        else
          if ((r-1)/2) mod 2 <> 0 then
            P := REDUCE4(1,r+1,nplus1,q);
            D := P*D;
            P := DIFF_2_S(1,r+1,nplus1,q);
            D := P*D;
            w := 2;
          else
            P := REDUCE4(3,r+1,nplus1,q);
            D := P*D;
            P := DIFF_2_S(3,r+1,nplus1,q);
            D := P*D;
            w := 0;
          fi; 
        fi; 
      else  
        if q mod 4 = 1 then
          if 1 < r then  
            P := SWR(2,r+1,nplus1);
            D := P*D;
            P := REDUCE2(3,r+1,nplus1,q);
            D := P*D;
          fi;
          w := 0;
        else 
          if ((r-1)/2) mod 2 <> 0 then
            P := SWR(4,r+1,nplus1);
            D := P*D;
            if 3 < r then
              P := REDUCE4(5,r+1,nplus1,q);
            else
              P := IdentityMat(nplus1,gf);
            fi;
            b := primroot^(LogFFE(-v, primroot)/2);
            P[3][3] := (1/2)*one;
            P[4][3] := (1/2)*one;
            P[3][4] := -1/(2*b);
            P[4][4] := 1/(2*b);
            D := P*D;
            if 3 < r then
              P := DIFF_2_S(5,r+1,nplus1,q);
              D := P*D;
            fi;
            w := 0;
          else
            P := SWR(2,r+1,nplus1);
            D := P*D;
            if 1 < r then
              P := REDUCE4(3,r+1,nplus1,q);
            else
              P := IdentityMat(nplus1,gf);
            fi;  
            b := primroot^(LogFFE(-v,primroot)/2);
            P[1][1] := 1/2;
            P[2][1] := 1/2;
            P[1][2] := -1/(2*b);
            P[2][2] := 1/(2*b);
            D := P*D;              
            if 1 < r then
              P := DIFF_2_S(3,r+1,nplus1,q);
              D := P*D;
            fi;
            w := 2;
          fi; 
        fi; 
      fi; 
    elif r <> 0 then 
      w := 1;
      if q mod 4 = 1 then
        P := REDUCE2(2,r+1,nplus1,q);
        D := P*D;
      else
        if r mod 4 = 0 then
          P := REDUCE4(2,r+1,nplus1,q);
          D := P*D;
          P := DIFF_2_S(2,r+1,nplus1,q);
          D := P*D;
        else
          if 3 < r then
            P := REDUCE4(4,r+1,nplus1,q);
          else
            P := IdentityMat(nplus1, gf);
          fi;
          dummy := SUM_OF_SQUARES(-1,q);
          c := dummy[1];
          d := dummy[2];
          P[1][1] := c;
          P[1][3] := d;
          P[3][1] := d;
          P[3][3] := -c;
          D := P*D;
          P := DIFF_2_S(2,r+1,nplus1,q);
          D := P*D;
          P := IdentityMat(nplus1, gf);
          i := 3;
          while i <= r + 1 do
            P[i][i] := -one;
            i := i + 2;
          od;
          D := P*D;
        fi; 
      fi; 
    else 
      w := 1;
    fi; 

    return [D,r,w];
end);

InstallMethod(BaseChangeOrthogonalQuadratic, [ IsMatrix and IsFFECollColl, IsField and IsFinite ],
    function(mat, gf)  
    local A,r,w,row,dummy,i,j,h,D,P,t,a,b,c,d,e,s,
      zeros,posr,posk,nplus1,control,q,zero,one,n;
    # n is the projective dimension
    nplus1 := Size(mat);
    n := nplus1-1;
    r := nplus1;   
    q := Size(gf);
    row := 1;
    zero := Zero(gf);
    one := One(gf);
    A := List(mat, ShallowCopy);
    D := IdentityMat(nplus1, gf);
    zeros := [];
    for i in [1..nplus1] do
      zeros[i] := zero;
    od;      
    h := Length(Factors(q));
    while row + 2 <= r do
      if not IsZero( A[row][row] ) then
        control := false;
        i := row + 1;
        # check on the main diagonal; we look for a zero
        while i <= r and not control do
          if IsZero( A[i][i] ) then
             control := true;
          else
             i := i + 1;
          fi;
        od;

        # if there is a zero somewhere, then we go and get it.

        if control then
          P := SWR(row,i,nplus1);
          A := P*A*P;   
          D := P*D;
          A := RESET(A,nplus1,q);
          
        # Otherwise: look in other places.

        else
          dummy := true;
          i := row;
          while i <= r - 1 and dummy do
            j := i + 1;
            while j <= r and dummy do
              if not IsZero( A[i][j] ) then
                posr := i;
                posk := j;   
                dummy := false;
              else
                j := j + 1;
              fi; 
            od;
            i := i + 1;
          od;

          # If all is zero, STOP
          if dummy then
            P := IdentityMat(nplus1, gf);
            t := SQRT2(A[row][row],q);
            P[row][row] := 1/t;
            for i in [row + 1..r] do
              P[i][row] := SQRT2(A[i][i],q)/t;
            od;
            D := P*D;
            # Permutation of the variables, it is a parabolic
            r := row;
            P := PERM_VAR(nplus1,r);
            D := P*D;
            w := 1;
            r := r - 1;
            return [D,r,w];
          # Otherwise: A basischange
          else
            if IsZero( A[row+1][row+2] ) then
               if posr = row + 1 then
                  P := SWR(posk,row+2,nplus1);
               elif posk = row + 2 then
                  P := SWR(posr,row+1,nplus1);
               elif posr = row + 2 then
                  P := TransposedMat(PermutationMat((posk,posr,row+1),nplus1));
               else
                  P := SWR(posr,row+1,nplus1)*SWR(posk,row+2,nplus1);
               fi;
               A := P*A*TransposedMat(P);
               D := P*D;
               A := RESET(A,nplus1,q);
            fi;
            #A[row+1][row+2] <> 0
            P := IdentityMat(nplus1,gf);
            t := A[row+1][row+2];
            P[row][row+2] := A[row][row+1]/t;
            P[row+2][row+2] := 1/t;  
            if row + 3 <= nplus1 then
              for i in [row+3..nplus1] do
                P[i][row+2] := A[row+1][i]/t;
              od;
            fi;
            A := P*A*TransposedMat(P);   
            A := RESET(A,nplus1,q);
            D := P*D;
            # A has now that special form a_11*X_1^2+X_1*X_2 + G(X_0,X_2,...,X_n);
            b := A[row][row];
            P :=  IdentityMat(nplus1, gf);
            t :=  SQRT2(b/A[row+1][row+1],q);
            P[row][row+1] := t;
            A := P*A*TransposedMat(P);
            A := RESET(A,nplus1,q);
            D := P*D;
            #A [row][row] is now 0 
          fi;
        fi; 
      fi; 
      # check for zero row
      dummy := true;
      i := row + 1;
      while  (dummy and i <= nplus1) do
        if not IsZero( A[row][i] ) then
           dummy := false;
        else
           i := i + 1;
        fi;
      od;
      posk := i;
      # A is a zero row then...
      if dummy then
        P := IdentityMat(nplus1);
        P[row][row] := 0;
        P[r][row] := 1;
        for i in [row+1..r] do
          P[i][i] := 0;
          P[i-1][i] := 1;
        od;
        A := P*A*TransposedMat(P);
        D := P*D;
        r := r - 1;
      else
        if posk <> row + 1 then
          P := SWR(posk,row+1,nplus1);
          D := P*D;
          A := P*A*TransposedMat(P);
          A := RESET(A,nplus1,q);
        fi;
        # Now A[k][k+1] <> 0
        P := IdentityMat(nplus1, gf);
        t := A[row][row+1];
        P[row+1][row+1] := P[row+1][row+1]/t;
        for i in [row+2..nplus1] do
            P[i][row+1] := A[row][i]/t;
        od;
        D := P*D;
        A := P*A*TransposedMat(P);
        A := RESET(A,nplus1,q);  
        P := IdentityMat(nplus1,GF(q));
        for i in [row+1..nplus1] do
          P[i][row] := A[row+1][i];
        od;
        D := P*D;
        A := P*A*TransposedMat(P);
        A := RESET(A,nplus1,q);
        row := row + 2;
      fi;  
    od;
    # Now there can be at most two variables left.
    # Case by case:
    
    if r = row then
       if IsZero(A[row][row]) then
          r := r - 1;
          w := 2;
       else
          t := SQRT2(A[r][r],q);
          P := IdentityMat(nplus1, gf);
          P[r][r] := 1/t;
          D := P*D;
          P := PERM_VAR(nplus1,r);
          D := P*D;
          w := 1;
       fi;  
    else
       a := A[row][row];   
       b := A[row][row+1];  
       c := A[row+1][row+1];
       t := zero; 
       if a = t then
          if b = t then
             if c = t then 
             r := r - 2;
             w := 2;
          else
             P := IdentityMat(nplus1, gf);
             P[r][r] := 1/SQRT2(c,q);
             D := P*D;
             P := PERM_VAR(nplus1,r); 
             D := P*D;
             r := r - 1;
             w := 1;
          fi;
        else
          if c = t then
            P := IdentityMat(nplus1,GF(q));
            P[r][r] := 1/b;
            D := P*D;
          else
            P := IdentityMat(nplus1,GF(q));
            P[r-1][r-1] := 1/b;
            P[r][r-1] := c/b;
            D := P*D;
          fi;
          w := 2;
        fi;
      else #a <> t
        if b = t then   
          if c = t then
            P := IdentityMat(nplus1, gf);
            P[r-1][r-1] := 1/SQRT2(a,q);
            D := P*D;  
            P := PERM_VAR(nplus1,r-1);
            D := P*D;
            r := r - 1; 
          else
            P := IdentityMat(nplus1, gf);
            P[r-1][r-1] := 1/SQRT2(a,q);
            P[r][r-1] := SQRT2(c,q)/SQRT2(a,q);
            D := P*D;
            P := PERM_VAR(nplus1,r-1);
            D := P*D;
            r := r - 1;
          fi;
          w := 1;
        else
          if c = t then
            P := IdentityMat(nplus1, gf);
            P[r-1][r] := a/b;
            P[r][r] := 1/b;
            D := P*D;
            w := 2;
          else
            d := (a*c)/(b^2);
            if Trace(GF(q),d) = t then
              e := SQRT2(a,q);
              P := IdentityMat(nplus1, gf);
              s := QUAD_EQ(d,q,h);
              P[r-1][r-1] := (s+one)/e;
              P[r-1][r] := e/b;
              P[r][r-1] := s/e;
              P[r][r] := e/b;
              D := P*D;
              w := 2;
            else
              P := IdentityMat(nplus1, gf);
              P[r-1][r-1] := SQRT2(c,q)/b;
              P[r][r] := 1/SQRT2(c,q);
              D := P*D;
              if r > 2 then
                P := SWR(2,r-1,nplus1)*SWR(1,r,nplus1);
              else
                P := SWR(1,2,nplus1);
              fi;
              D := P*D;
              e := C1(q,h);
              if e <> d then
                 a := QUAD_EQ(d+e,q,h);
                 P := IdentityMat(nplus1, gf);
                 P[2][1] := a; 
                 D := P*D;
              fi;
              w := 0;
            fi;   
          fi;
        fi;
      fi;
    fi;
    r := r - 1;    
    return [D,r,w];
end );
      
InstallMethod(BaseChangeHermitian, [ IsMatrix and IsFFECollColl, IsField and IsFinite ],
  function(mat,gf)
    local row,i,j,k,A,a,b,P,D,t,dummy,dummy2,r,stop,nplus1,q,one,zero,n;
    A := mat;
    n := Size(mat) - 1; # projective dimension
    nplus1 := n + 1;
    one := One(gf);
    zero := Zero(gf);
    q := Size(gf);
    D := IdentityMat(nplus1, gf);
    row := 0;
    stop := false;
    t := Sqrt(q);
    repeat
      # We look for a nonzero element on the main diagonal, starting
      # from row + 1
      i := 1 + row;
      dummy := false;
      while dummy = false and i <= nplus1 do
        if IsZero( A[i][i] ) then
          i := i + 1;
        else
           dummy := true;
        fi;
      od;
      # if i = row + 1 then we do nothing, since A[i][i] <> 0
      if 2 + row <= i and i <= nplus1 then
        P := SWR(row + 1,i,nplus1);
        A := P*A*P;  #HERM_CONJ(P) = P, so HERM_CONJ is not necessary
        D := P*D;

      # If we cannot find it, we got elsewhere to get it.

      elif i = n + 2 then
        i := 1 + row;
        dummy := false;
        while i <= n and dummy = false do
          k := i + 1; 
          while k <= nplus1 and dummy = false do
            if IsZero( A[i][k] ) then
               k := k + 1;
            else
               dummy := true;
            fi;
          od;
          if k = n + 2 then
             i := i + 1;
          fi;
        od;

        # Is i = n+1, then it is all zero an we can stop.

        if i = nplus1 then
          stop := true;
          r := row;
          # Otherwise, go and get it
          # Start from A[row+1,row+2]
        elif i = row + 1 then
          P := SWR(row+2,k,nplus1);
          A := P*A*P;
          D := P*D;
        else
          P := SWR(row+2,k,nplus1)*SWR(row+1,i,nplus1);
          A := P*A*TransposedMat(P);
          D := P*D;
        fi;
        
        if not stop then
          b := Z(q)*(A[row+2][row+1])^-1;
          P := IdentityMat(nplus1, gf);
          P[row+1][row+2] := b;
          A := P*A*HERM_CONJ(P,nplus1,t);
          D := P*D;
        fi;
      fi;    

      if not stop then
        P := IdentityMat(nplus1, gf);
        for i in [row+2..nplus1] do
           P[i][row+1] := -A[i][row+1]*(A[row+1][row+1])^-1;
        od;
        A := P*A*HERM_CONJ(P,nplus1,t);
        D := P*D;
        row := row + 1;
      fi;
    until row = n or stop;

   # Count how many variables have been used 

    if not stop then
      if not IsZero(A[nplus1][nplus1]) then
         r := nplus1;
      else
         r := n;
      fi;
    fi;
    r := r - 1;
    # Take care that the diagonal elements become 1.

    dummy := Difference(gf, GF(t));
    P := IdentityMat(nplus1, gf);
    for i in [1..r+1] do
      a := A[i][i];
      if not IsOne(a) then
        j := 1;
        b := dummy[j];
        while a <> b*b^t do
          j := j + 1;
          b := dummy[j];
        od;
        P[i][i] := 1/b;
      fi;
    od;
    D := P*D;
    return [D,r];
end );

InstallMethod( BaseChangeSymplectic, [ IsMatrix and IsFFECollColl, IsField and IsFinite ],

## This operation returns an isometry g such that g m g^T is
## the form arising from the block diagonal matrix
## with each block equal to J=[[0,1],[-1,0]]. 

  function( m, f )
    local c1, c2, pos, d, o, iter, basechange, alpha;
    d := Size(m);
    o := Zero(f);
    basechange := [];

    for pos in [1..d/2] do
        iter := Iterator(f^d);
        repeat
          c1 := NextIterator(iter);
        until not IsZero(c1) and not ForAny(basechange, x -> c1*m*x <> o);
      
        ## Find c2 which is orthogonal to c1
        iter := Iterator(f^d);
        repeat
          c2 := NextIterator(iter);
          alpha := c1*m*c2;
        until not IsZero(c2) and alpha <> o;

        ## normalise 
        c2 := c2 / alpha;
        Append(basechange, [c1,c2]);
    od;
    return basechange;
  end );

#############################################################################
# Other Operations:
#############################################################################
  
InstallMethod( BaseChangeHomomorphism, [ IsMatrix and IsFFECollColl, IsField ],
  function( b, gf )
  ## This function returns an intertwiner of the general linear group
  ## induced by changing the basis of its underlying vector space 

    local gl, gens, newgens, invb, newgl, hom;
    if IsZero(Determinant(b)) then
       Error("Matrix is not invertible");
    fi;
    invb := Inverse( b );
    gl := GeneralLinearGroup(Size(b), gf);
    hom := InnerAutomorphismNC( gl, invb); 
    return hom;
  end );

InstallMethod( BaseField, "for a form", [IsForm],
  function( f )
    return f!.basefield;
  end );

InstallMethod( GramMatrix, "for a form", [IsForm],
  function( f )
    return f!.matrix;
  end );

InstallMethod( IsDegenerateForm, [IsForm],
  function( f )
    return not IsTrivial( RadicalOfForm( f ) );
  end );

InstallMethod( IsReflexiveForm, [IsBilinearForm],
  function( f )
    ## simply check if form is symmetric or alternating
    local m;
    m := f!.matrix;
    return m = TransposedMat(m) or m = -TransposedMat(m);
  end );

InstallMethod( IsReflexiveForm, [IsHermitianForm],
  function( f )
    local m, frob;
    m := f!.matrix;
    frob := FrobeniusAutomorphism(f!.basefield);
    return m = TransposedMat(m^frob);
  end );

InstallMethod( IsAlternatingForm, [IsBilinearForm],
  function( f )
    return IsSymplecticMatrix( f!.matrix, f!.basefield );
 end );

InstallMethod( IsSymmetricForm, [IsBilinearForm],
  function( f )
    return IsOrthogonalMatrix( f!.matrix, f!.basefield );
 end );

InstallMethod( RadicalOfForm, [IsSesquilinearForm],
  function( f )
    local m, null, gf, d, rad;
    if not IsReflexiveForm( f ) then
       Error( "Form must be reflexive\n");
    fi;
    m := f!.matrix;
    gf := f!.basefield;
    d := Size(m);
    null := NullspaceMat( m );
    return Subspace( gf^d, null );
  end );

InstallMethod( RadicalOfForm, [IsQuadraticForm],
  function( f )
    local m, null, gf, d;
    m := f!.matrix;
    m := m + TransposedMat(m);
    gf := f!.basefield;
    d := Size(m);
    null := NullspaceMat( m );
    return Subspace( gf^d, null );
  end );

InstallMethod( DiscriminantOfForm, [ IsQuadraticForm ],
 function( f )
   local m, gf, d, det, squares, primroot;
   m := f!.matrix;
   gf := f!.basefield;
   d := Size(m);
   if IsOddInt(d) then
      Error( "Quadratic form must be defined by a matrix of even dimension");
   fi;
   det := Determinant( m + TransposedMat(m) );
   if IsZero( det ) then
      Error( "Form must be nondegenerate" );
   fi;
   if IsEvenInt(Size(gf)) then
      return "square";
   fi;
   if IsOddInt(Size(gf)) then
      primroot := PrimitiveElement( gf );
      squares := AsList( Group( primroot^2 ) );
      if det in squares then
         return "square";
      else
         return "nonsquare";
      fi;
   else
      return "square";
   fi;
 end );

InstallMethod( DiscriminantOfForm, [ IsSesquilinearForm ],
 function( f )
   local m, gf, d, det, squares, primroot;
   m := f!.matrix;
   gf := f!.basefield;
   d := Size(m);
   if IsOddInt(d) then
      Error( "Sesquilinear form must be defined by a matrix of even dimension");
   fi;
   det := Determinant( m );
   if IsZero( det ) then
      Error( "Form must be nondegenerate" );
   fi;
   if IsEvenInt(Size(gf)) then
      return "square";
   fi;
   if IsOddInt(Size(gf)) then
      primroot := PrimitiveElement( gf );
      squares := AsList( Group( primroot^2 ) );
      if det in squares then
         return "square";
      else
         return "nonsquare";
      fi;
   else
      return "square";
   fi;
 end );

InstallMethod( WittIndex, "for a bilinear form",
  [IsBilinearForm and IsFormRep],
  function(f)
  local string;
  string := f!.type; 
  if string = "symplectic" then
     return Length(f!.matrix) / 2;
  elif string = "orthogonal" then
     if IsEllipticForm( f ) then
        return WittIndex( f );
     elif IsHyperbolicForm( f ) then
        return WittIndex( f );
     elif IsParabolicForm( f ) then
        return WittIndex( f );
     else
        Error("Form is incorrectly specified\n");
     fi;
  else 
    Error("Form is incorrectly specified\n");
  fi;
  end );

InstallMethod( WittIndex, "for a quadratic form",
  [IsQuadraticForm and IsFormRep],
  function(f)
  if IsEllipticForm( f ) then
     return WittIndex( f );
  elif IsHyperbolicForm( f ) then
     return WittIndex( f );
  elif IsParabolicForm( f ) then
     return WittIndex( f );
  else
     Error("Form is incorrectly specified\n");
  fi;
  end );

InstallMethod( WittIndex, "for an hermitian form",
  [IsHermitianForm and IsFormRep],
  function(f)
  local d;
  d := Length(f!.matrix);
  return Int( (d+1)/2 );
  end );

InstallMethod( IsSymplecticForm, "for sesquilinear forms",
  [IsSesquilinearForm and IsFormRep],
  function( f )
    local string;
    string := f!.type;
    if string = "symplectic" then
       if IsAlternatingForm( f ) then
          return true;
       else
          Error( "Form was incorrectly specified" );
       fi;
    else
       return false;
    fi;
  end );

InstallMethod( IsEllipticForm, "for sesquilinear forms",
  [IsSesquilinearForm and IsFormRep],
  function(f)
    local b,string,gf,m;
    string := f!.type;
    gf := f!.basefield;
    m := f!.matrix;
    if string = "orthogonal" then
      b := BaseChangeOrthogonalBilinear(m, gf);
      if not HasWittIndex(f) then
         SetWittIndex(f, (b[2]+b[3]-1) / 2);
      fi;
      return b[3] = 0;
    else
      return false;
    fi;
  end );

InstallMethod( IsEllipticForm,  "for quadratic forms",
  [IsQuadraticForm and IsFormRep],
  function(f)
    local b,string,gf,m;
    string := f!.type;
    gf := f!.basefield;
    m := f!.matrix;
    b := BaseChangeOrthogonalQuadratic(m, gf);
    if not HasWittIndex(f) then
       SetWittIndex(f, (b[2]+b[3]-1) / 2);
    fi;
    return b[3] = 0;
  end );

InstallMethod( IsParabolicForm, "for sesquilinear forms",
  [IsSesquilinearForm and IsFormRep],
  function(f)
  local b,string,gf,m;
    string := f!.type;
    gf := f!.basefield;
    m := f!.matrix;
    if string = "orthogonal" then
      b := BaseChangeOrthogonalBilinear(m, gf);
      if not HasWittIndex(f) then
         SetWittIndex(f, (b[2]+b[3]-1) / 2);
      fi;
      return b[3] = 1;
    else
      return false;
    fi;
  end );

InstallMethod( IsParabolicForm, "for quadratic forms",
  [IsQuadraticForm and IsFormRep],
  function(f)
  local b,string,gf,m;
    string := f!.type;
    gf := f!.basefield;
    m := f!.matrix;
    b := BaseChangeOrthogonalQuadratic(m, gf);
    if not HasWittIndex(f) then
       SetWittIndex(f, (b[2]+b[3]-1) / 2);
    fi;
    return b[3] = 1;
  end );

InstallMethod( IsHyperbolicForm, "for sesquilinear forms",
  [IsSesquilinearForm and IsFormRep],
  function(f)
    local b,string,gf,m;
    string := f!.type;
    gf := f!.basefield;
    m := f!.matrix;
    if string = "orthogonal" then
      b := BaseChangeOrthogonalBilinear(m, gf);
      if not HasWittIndex(f) then
         SetWittIndex(f, (b[2]+b[3]-1) / 2);
      fi;
      return b[3] = 2;
    else
      return false;
    fi;
end );

InstallMethod( IsHyperbolicForm, "for quadratic forms",
  [IsQuadraticForm and IsFormRep],
  function(f)
  local b,string,gf,m;
  string := f!.type;
  gf := f!.basefield;
  m := f!.matrix;
  b := BaseChangeOrthogonalQuadratic(m, gf);
  if not HasWittIndex(f) then
     SetWittIndex(f, (b[2]+b[3]-1) / 2);
  fi;
  return b[3] = 2;
end );

InstallMethod( IsometricCanonicalForm, "for sesquilinear forms",
  [IsBilinearForm and IsFormRep],
  function(f)
  local B,gram,isom,gf,form;
  gram := GramMatrix(f);
  B := BaseChangeToCanonical(f);
  isom := B*gram*TransposedMat(B);
  gf := f!.basefield;
  form := FormByMatrix(isom,gf,"orthogonal");
  SetWittIndex(form, WittIndex(f));
  return form;
end );

InstallMethod( IsometricCanonicalForm, "for hermitian forms",
  [IsHermitianForm and IsFormRep],
  function(f)
  local B,gram,isom,gf,form;
  gram := GramMatrix(f);
  B := BaseChangeToCanonical(f);
  gf := f!.basefield;
  isom := B*gram*HERM_CONJ(B,Length(B),Sqrt(Size(gf))); 
      #do not forget that HERM_CONJ also transposes
  form := FormByMatrix(isom,gf,"hermitian");
  SetWittIndex(form, WittIndex(f));
  return form;
end );

InstallMethod( IsometricCanonicalForm, "for quadratic forms",
  [IsQuadraticForm and IsFormRep],
  function(f)
  local B,gram,isom,gf,form;
  gram := GramMatrix(f);
  gf := f!.basefield;
  B := BaseChangeToCanonical(f);
  isom := RESET(B*gram*TransposedMat(B),Length(gram),Size(gf));
  form := FormByMatrix(isom,gf,"quadratic");
  SetWittIndex(form, WittIndex(f));
  return form;
end );

#############################################################################
# Evaluation Methods: 
#############################################################################

InstallMethod( EvaluateForm,  "for a bilinear form and a pair of vectors",
  [IsBilinearForm and IsFormRep, 
        IsVector and IsFFECollection, IsVector and IsFFECollection],
  function(f,v,w)
  return v*f!.matrix*w;
end );

InstallMethod( EvaluateForm,  "for a bilinear form and a pair of matrices",
  [IsBilinearForm and IsFormRep, IsFFECollColl, IsFFECollColl],
  function(f,v,w)
  return v*f!.matrix*TransposedMat(w);
end );

InstallMethod( EvaluateForm, "for an hermitian form and a pair of vectors",
  [IsHermitianForm and IsFormRep,         
        IsVector and IsFFECollection, IsVector and IsFFECollection],
  function(f,v,w)
  local gf,t;
  gf := f!.basefield;
  t := Sqrt(Size(gf));
  return v*f!.matrix*(w^t);
end );

InstallMethod( EvaluateForm,  "for an hermitian form and a pair of matrices",
  [IsHermitianForm and IsFormRep, IsFFECollColl, IsFFECollColl],
  function(f,v,w)
  local gf,frob;
  gf := f!.basefield;
  frob := FrobeniusAutomorphism(gf);
  return v*f!.matrix*TransposedMat(w^frob);
end );

InstallMethod( EvaluateForm, "for quadratic forms",
  [IsQuadraticForm and IsFormRep, IsVector and IsFFECollection],
  function(f,v)
  return v*f!.matrix*v;
end );

